// Copyright (c) 2017, Tobias Mueller tm(at)tm3d.de
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the
//    distribution.
//  * All advertising materials mentioning features or use of this
//    software must display the following acknowledgement: This product
//    includes software developed by tm3d.de and its contributors.
//  * Neither the name of tm3d.de nor the names of its contributors may
//    be used to endorse or promote products derived from this software
//    without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include "owUSBInterface.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <pthread.h>
#include <memory.h>




extern "C" {
#include "libusbds2490.h"
}




namespace usb_1w {
	
	
#define FALSE 0
#define TRUE 1
int usb_num_devices = -1;
usb_dev_handle *usb_dev_handle_list[20];
//static SMALLINT usbhnd_init = 0;
struct usb_device *usb_dev_list[20];
static void DS2490Discover(void)
{
	struct usb_bus *bus;
	struct usb_device *dev;

	// initialize USB subsystem
#ifdef LIBUSB_DEBUG_DISABLE  // Use this define at compile time to turn off LIBUSB debug info
   usb_set_debug(0);
#endif
   usb_init();
   usb_find_busses();
   usb_find_devices();

   usb_num_devices = 0; // avoid port zero to make it look like other builds

	for (bus = usb_get_busses(); bus; bus = bus->next) 
	{
				
		for (dev = bus->devices; dev; dev = dev->next) 
		{
			//printf("%04X:%04X\n",dev->descriptor.idVendor,dev->descriptor.idProduct);
			if (dev->descriptor.idVendor == 0x04FA &&
				 dev->descriptor.idProduct == 0x2490) 
			{
				//printf("DS2490Discover-find\n");
				usb_dev_list[++usb_num_devices] = dev;
				
			}
		}
	}
}


uint8_t AdapterRecover(usb_dev_handle *hDevice)
{
   // detect DS2490 and set it up
   if (DS2490Detect(hDevice))
   {
      //USBSpeed[portnum] = MODE_NORMAL; // current DS2490 1-Wire Net communication speed
      //USBLevel[portnum] = MODE_NORMAL; // current DS2490 1-Wire Net level
      return TRUE;
   }
   else
   {
      //OWERROR(OWERROR_SYSTEM_RESOURCE_INIT_FAILED);
      return FALSE;
   }
}

}

using namespace usb_1w;
int id[8];
int first=1;

int owUSBInterface::InitAdapter(uint8_t pin){
    if(!usbhnd_init)  // check to see the USB bus has been setup properly
    {
	    // discover all DS2490s on USB bus...
	    DS2490Discover();
        usbhnd_init = 1;
    }
	if (usb_num_devices==0) return -1;

    // check to see if the portnumber is valid
    if ((usb_num_devices <pin) || (pin == 0)) 
    {
		// error - Attempted to select invalid port number
        log->set(OWLOG_ERROR,"OWERROR_LIBUSB_NO_ADAPTER_FOUND");
	    return -1;
    }

    // check to see if opening the device is valid
    if (hDevice!= NULL) 
	{
		// error - USB Device already open
		log->set(OWLOG_ERROR,"OWERROR_LIBUSB_DEVICE_ALREADY_OPENED");
        return -1;
    }

    // open the device
    hDevice = usb_open(usb_dev_list[pin]);
    if (hDevice == NULL) 
    {
        // Failed to open usb device
		log->set(OWLOG_ERROR,"OWERROR_LIBUSB_OPEN_FAILED");
		return -1;
    }

    // set the configuration
    if (usb_set_configuration(hDevice, 1)) 
    {
        // Failed to set configuration
		log->set(OWLOG_ERROR,"OWERROR_LIBUSB_SET_CONFIGURATION_ERROR");
        usb_close(hDevice); // close handle
	    return -2;
    }

    // claim the interface
    if (usb_claim_interface(hDevice, 0)) 
    {
        // Failed to claim interface
	    //printf("%s\n", usb_strerror());
		log->set(OWLOG_ERROR,"OWERROR_LIBUSB_CLAIM_INTERFACE_ERROR");
	    usb_close(hDevice); // close handle
		return -1;
	}

	// set the alt interface
	if (usb_set_altinterface(hDevice, 3)) 
	{
		// Failed to set altinterface
		//printf("%s\n", usb_strerror());
		log->set(OWLOG_ERROR,"OWERROR_LIBUSB_SET_ALTINTERFACE_ERROR");
		usb_release_interface(hDevice, 0); // release interface
		usb_close(hDevice);   // close handle
		return -1;
	}

   
    // clear USB endpoints before doing anything with them.
	usb_clear_halt(hDevice, DS2490_EP3);
	usb_clear_halt(hDevice, DS2490_EP2);
	usb_clear_halt(hDevice, DS2490_EP1);

    // verify adapter is working
    if (!AdapterRecover(hDevice))
    {
		log->set(OWLOG_ERROR,"Adapter don't work");
		usb_release_interface(hDevice, 0); // release interface
		usb_close(hDevice);   // close handle
        return -1;
    }

    // reset the 1-Wire
    Reset();

    return 1;
 }
int owUSBInterface::Reset() {
   SETUP_PACKET setup;
   SMALLINT present,vpp;
   SMALLINT ret = 0;

   // construct command
   setup.RequestTypeReservedBits = 0x40;
   setup.Request = COMM_CMD;
   setup.Value = COMM_1_WIRE_RESET | COMM_F | COMM_IM | COMM_SE;
   setup.Index =  ONEWIREBUSSPEED_REGULAR;// ONEWIREBUSSPEED_FLEXIBLE;
   setup.Length = 0;
   setup.DataOut = FALSE;
   // call the libusb driver
   ret = usb_control_msg(hDevice, 
                         setup.RequestTypeReservedBits, 
			 setup.Request, 
			 setup.Value, 
			 setup.Index, 
			 NULL, 
			 setup.Length, 
			 TIMEOUT_LIBUSB);
   if (ret < 0)
   {
      // failure
      log->set(OWLOG_ERROR,"OWERROR_RESET_FAILED");
      //AdapterRecover(portnum);
      return FALSE;
   }
   else
   {
      // extra delay for alarming DS1994/DS2404 complience
     // if ((FAMILY_CODE_04_ALARM_TOUCHRESET_COMPLIANCE) && (USBSpeed[portnum] != MODE_OVERDRIVE))
     //    msDelay(5);

      // success, check for shorts
      if (DS2490ShortCheck(hDevice, &present,&vpp))
      {
         //USBVpp[portnum] = vpp;     
         return present;
      }
      else
      {
       log->set(OWLOG_ERROR,"OWERROR_OW_SHORTED");
         // short occuring
         usleep(300);
         //AdapterRecover(portnum);
         return FALSE;
      }
   }
}


uint8_t owUSBInterface::sendrecivBit(uint8_t bit) {
	SETUP_PACKET setup;
	ushort nBytes;   
	uchar buf[2];
	SMALLINT ret = 0;

	// set to do touchbit
	setup.RequestTypeReservedBits = 0x40;
	setup.Request = COMM_CMD;
	setup.Value = COMM_BIT_IO | COMM_IM | ((bit) ? COMM_D : 0);
	setup.Index = 0;  
	setup.Length = 0;
	setup.DataOut = FALSE;

   // call the libusb driver
   ret = usb_control_msg(hDevice, 
                         setup.RequestTypeReservedBits, 
			 setup.Request, 
			 setup.Value, 
			 setup.Index, 
			 NULL, 
			 setup.Length, 
			 TIMEOUT_LIBUSB);
   if (ret < 0)
   {
      // failure
      log->set(OWLOG_ERROR,"OWERROR_ADAPTER_ERROR");
      AdapterRecover(hDevice);
      return 0;
   }
   else
   {
      // success, read the result
      nBytes = 1;
      if (DS2490Read(hDevice, buf, &nBytes)) {
		usleep(70);
         return buf[0];
	}
      else
      {
         log->set(OWLOG_ERROR,"OWERROR_ADAPTER_ERROR");
         AdapterRecover(hDevice);
         return 0;
      }
   }
}

uint8_t owUSBInterface::sendrecivByte(uint8_t byte) {
	SETUP_PACKET setup;
    ushort nBytes;   
	uchar buf[2];
	SMALLINT ret = 0;

	// set to do touchbyte
	setup.RequestTypeReservedBits = 0x40;
	setup.Request = COMM_CMD;
	setup.Value = COMM_BYTE_IO | COMM_IM;
	setup.Index = byte & 0xFF;  
	setup.Length = 0;
	setup.DataOut = FALSE;

	// call the libusb driver
	ret = usb_control_msg(hDevice, 
	                 setup.RequestTypeReservedBits, 
			 setup.Request, 
			 setup.Value, 
			 setup.Index, 
			 NULL, 
			 setup.Length, 
			 TIMEOUT_LIBUSB);
   if (ret < 0)
   {
      // failure
      log->set(OWLOG_ERROR,"OWERROR_ADAPTER_ERROR");
      AdapterRecover(hDevice);
      return 0;
   }
   else
   {
      // success, read the result
      nBytes = 1;

      if (DS2490Read(hDevice, buf, &nBytes)) {
         usleep(80);
         return buf[0];
         }
      else
      {
         log->set(OWLOG_ERROR,"OWERROR_ADAPTER_ERROR");
         AdapterRecover(hDevice);
         return 0;
      }
   }
}

void owUSBInterface::ReleaseAdapter() {
	usb_release_interface(hDevice, 0);
	usb_close(hDevice);
	hDevice = NULL;
}
